#!/usr/bin/env bash

declare -r SCRIPTNAME="$(basename "$0")"
declare -r SCRIPTVERSION="1.0"

# script configuration
declare -r SubdirName='overburden'
declare -r DropInConfigurationName='use_nooverburden_geometry_icarus.fcl'


# ------------------------------------------------------------------------------
function help() {
  cat <<EOH
Creates a no-overburden version of the specified original job file.

Usage:  ${SCRIPTNAME} OriginalFHiCLfile NewFHiCLname

The new file includes OriginalFHiCLfile and it is placed in the \`${SubdirName}\` subdirectory
of the directory that already contains it (or the current one if path is not
included in OriginalFHiCLfile).

EOH

} # help()


function printUserName() {

  local UserName
  local Email
  
  # try to get the information from GIT
  local GitDirectory
  [[ -d ".git" ]] && GitDirectory="$(pwd)"
  [[ -d "${MRB_SOURCE}/icaruscode/.git" ]] && GitDirectory="${MRB_SOURCE}/icaruscode"
  if [[ -d "$GitDirectory" ]]; then
    UserName="$(git -C "$GitDirectory" config --get user.name 2> /dev/null)"
    Email="$(git -C "$GitDirectory" config --get user.email 2> /dev/null)"
  fi
  
  # fallback: ask the operating system
  [[ -z "$UserName" ]] && UserName="$USER"
  [[ -z "$UserName" ]] && UserName="$(whoami)"
  
  echo "${UserName}${Email:+" (${Email})"}"
} # printUserName()


# ------------------------------------------------------------------------------

if [[ $# -lt 2 ]]; then
  help
  exit
fi

declare -r SourceFile="$1"
declare -r DestName="$(basename "${2%.fcl}.fcl")"

declare -r SourceDir="$(dirname "$SourceFile")"
declare -r SourceName="$(basename "${SourceFile%.fcl}.fcl")"
declare -r DestDir="${SourceDir%/}/${SubdirName}"
declare -r DestFile="${DestDir:+"${DestDir%/}/"}${DestName}"

if [[ "$SourceName" == "$DestName" ]]; then
  echo "Cowardly refusing to create a new FHiCL file with the same name as the original one ('${SourceName}')." >&2
  exit 1
fi

echo "'${SourceName}' => '${DestName}'"

[[ -d "$DestDir" ]] || mkdir "$DestDir"

cat <<EOF > "$DestFile"
#
# File:    ${DestName}
# Purpose: No-overburden version of \`${SourceName}\`.
# Author:  $(printUserName)
# Date:    $(date)
# 
# This file was automatically generated by ${SCRIPTNAME} version ${SCRIPTVERSION}.
#

#include "${SourceName}"

# turn to no-overburden geometry:
#include "${DropInConfigurationName}"

EOF
